from typing import List, NoReturn
from pymarc.record import Record
from rara_tools.parsers.marc_records.base_record import BaseRecord
from rara_tools.constants.parsers import OrganizationMarcIDs, LOGGER
import regex as re
import json

# TODO: indikaatorid ind1 väljadel 100 ja 400?


class OrganizationRecord(BaseRecord):
    """ Generates a simplified organization JSON record
    from a pymarc MARC record.
    """
    def __init__(self, record: Record, add_variations: bool = False) -> NoReturn:
        """ Initializes OrganizationRecord object.

        Parameters
        -----------
        record: Record
            pymarc.record.Record object.
        add_variations: bool
            If enabled, constructs an additional variations field, which
            combines the content of multiple fields + adds some generated
            variations. If the output is uploaded into Elastic and used
            via rara-norm-linker, it is necessary to enable this.
        """
        super().__init__(record=record, add_variations=add_variations)

        self.__name_field_id: List[str] = OrganizationMarcIDs.NAME
        self.__name_variations_field_id: List[str] = OrganizationMarcIDs.NAME_VARIATIONS
        self.__related_names_field_id: List[str] = OrganizationMarcIDs.RELATED_NAMES
        self.__source_field_id: List[str] = OrganizationMarcIDs.SOURCE
        self.__description_field_id: List[str] = OrganizationMarcIDs.DESCRIPTION
        self.__area_code_id: List[str] = OrganizationMarcIDs.AREA_CODE
        self.__default_year: int | None = None

        self.__name: str = ""
        self.__original_name: dict = {}
        self.__name_specification: str = ""
        self.__dates: str = ""
        self.__location: str = ""
        self.__numeration: str = ""
        self.__name_variations: List[str] = []
        self.__source: str = ""
        self.__description: str = ""
        self.__area_code: str = ""
        self.__acronyms: List[str] = []
        self.__alternative_names: List[str] = []
        self.__related_acronyms: List[str] = []
        self.__old_names: List[str] = []
        self.__new_names: List[str] = []
        self.__related_old_names: List[str] = []
        self.__related_new_names: List[str] = []
        self.__full_record: dict = {}
        self.__variations: List[str] = []


    def _clean_value(self, value: str) -> str:
        try:
            cleaned_value = value.strip("., ")
        except Exception as e:
            cleaned_value = ""
        return cleaned_value

    def _merge_and_clean(self, value: dict, keys: List[str]) -> str:
        _merged = []
        for key in keys:
            _value = self._clean_value(value.get(key, ""))
            if _value:
                _merged.append(_value)
        merged = " ".join(_merged)
        return merged

    @property
    def original_name(self) -> str:
        if not self.__original_name:
            values = self.get_values(
                marc_ids=self.__name_field_id,
                subfield_id=["a", "b"]
            )
            if values:
                self.__original_name = {
                    "a": self._clean_value(values[0].get("a", "")),
                    "b": self._clean_value(values[0].get("b", ""))
                }
            else:
                LOGGER.info(
                    f"Could not parse subfields 'a' and/or 'b' from " \
                    f"field {self.__name_field_id}. Record:\n{self.marc_record}"
                )
        return self.__original_name

    @property
    def name(self) -> str:
        if not self.__name:
            self.__name = self._merge_and_clean(self.original_name, ["a", "b"])
        return self.__name

    @property
    def dates(self) -> str:
        if not self.__dates:
            values = self.get_values(
                marc_ids=self.__name_field_id,
                subfield_id="d"
            )
            if values:
                self.__dates = self._clean_value(values[0])
        return self.__dates

    @property
    def location(self) -> str:
        if not self.__location:
            values = self.get_values(
                marc_ids=self.__name_field_id,
                subfield_id="c"
            )
            if values:
                self.__location = self._clean_value(values[0])
        return self.__location

    @property
    def numeration(self) -> str:
        if not self.__numeration:
            values = self.get_values(
                marc_ids=self.__name_field_id,
                subfield_id="n"
            )
            if values:
                self.__numeration = self._clean_value(values[0])
        return self.__numeration

    @property
    def acronyms(self) -> List[str]:
        if not self.__acronyms:
            values = self.get_values(
                marc_ids=self.__name_variations_field_id,
                subfield_id="a",
                subfield_restriction = ("w", "d")
            )
            self.__acronyms = [self._clean_value(value) for value in values]
        return self.__acronyms

    @property
    def new_names(self) -> List[str]:
        if not self.__new_names:
            values = self.get_values(
                marc_ids=self.__name_variations_field_id,
                subfield_id=["a", "b"],
                subfield_restriction = ("w", "b")
            )
            self.__new_names = [self._merge_and_clean(value, ["a", "b"]) for value in values]
        return self.__new_names

    @property
    def old_names(self) -> List[str]:
        if not self.__old_names:
            values = self.get_values(
                marc_ids=self.__name_variations_field_id,
                subfield_id=["a", "b"],
                subfield_restriction = ("w", "a")
            )
            self.__old_names = [self._merge_and_clean(value, ["a", "b"]) for value in values]
        return self.__old_names

    @property
    def alternative_names(self) -> List[str]:
        if not self.__alternative_names:
            values = self.get_values(
                marc_ids=self.__name_variations_field_id,
                subfield_id=["a", "b"],
                subfield_to_ignore="w"
            )
            self.__alternative_names = [self._merge_and_clean(value, ["a", "b"]) for value in values]
        return self.__alternative_names


    @property
    def related_acronyms(self) -> List[str]:
        if not self.__related_acronyms:
            values = self.get_values(
                marc_ids=self.__related_names_field_id,
                subfield_id="a",
                subfield_restriction = ("w", "d")
            )
            self.__related_acronyms = [self._clean_value(value) for value in values]
        return self.__related_acronyms

    @property
    def related_new_names(self) -> List[str]:
        if not self.__related_new_names:
            values = self.get_values(
                marc_ids=self.__related_names_field_id,
                subfield_id=["a", "b"],
                subfield_restriction = ("w", "b")
            )
            self.__related_new_names = [self._merge_and_clean(value, ["a", "b"]) for value in values]
        return self.__related_new_names

    @property
    def related_old_names(self) -> List[str]:
        if not self.__related_old_names:
            values = self.get_values(
                marc_ids=self.__related_names_field_id,
                subfield_id=["a", "b"],
                subfield_restriction = ("w", "a")
            )
            self.__related_old_names = [self._merge_and_clean(value, ["a", "b"]) for value in values]
        return self.__related_old_names


    @property
    def source(self) -> str:
        if not self.__source:
            values = self.get_values(
                marc_ids=self.__source_field_id,
                subfield_id="a"
            )
            self.__source = self._clean_value(values[0]) if values else ""
        return self.__source


    @property
    def area_code(self) -> str:
        if not self.__area_code:
            values = self.get_values(
                marc_ids=self.__area_code_id,
                subfield_id="c"
            )
            self.__area_code = self._clean_value(values[0]) if values else ""
        return self.__area_code

    @property
    def description(self) -> str:
        if not self.__description:
            values = self.get_values(
                marc_ids=self.__description_field_id,
                subfield_id="i"
            )
            self.__description = self._clean_value(values[0]) if values else ""
        return self.__description

    @property
    def variations(self) -> List[str]:
        if not self.__variations:
            _variations = [self.name]
            _variations.extend(self.new_names)
            _variations.extend(self.old_names)
            _variations.extend(self.alternative_names)
            _variations.extend(self.related_old_names)
            _variations.extend(self.related_new_names)
            self.__variations = [v.lower() for v in list(set(_variations))]

        return self.__variations

    @property
    def full_record(self) -> dict:
        if not self.__full_record:
            self.__full_record = {
                "name": self.name,
                "original_name": self.original_name,
                "acronyms": self.acronyms,
                "new_names": self.new_names,
                "old_names": self.old_names,
                "source": self.source,
                "description": self.description,
                "area_code": self.area_code,
                "alternative_names": self.alternative_names,
                "related_acryonyms": self.related_acronyms,
                "related_new_names": self.related_new_names,
                "related_old_names": self.related_old_names,
                "identifier": self.identifier,
                "identifier_source": self.identifier_source,
                "full_record_marc": str(self.marc_record),
                "full_record_json": json.dumps(self.marc_json_record)
            }
            if self.add_variations:
                self.__full_record.update(
                    {
                        "link_variations": self.variations,
                        "link_acronyms": [a.lower() for a in self.acronyms]
                    }
                )
        return self.__full_record



from datetime import datetime
from pymarc import (Field, Subfield, JSONReader, Record)
from typing import List, Optional, Iterator
from rara_tools.normalizers.reader import SafeJSONReader

from rara_tools.parsers.tools.validators import filter_names

from rara_tools.normalizers.viaf import VIAFRecord, VIAFClient
from rara_tools.constants.normalizers import (
    DEFAULT_VIAF_FIELD, ALLOWED_VIAF_FIELDS, ALLOWED_VIAF_WIKILINK_LANGS,
    VIAF_SIMILARITY_THRESHOLD, VERIFY_VIAF_RECORD, MAX_VIAF_RECORDS_TO_VERIFY,
    EMPTY_INDICATORS, YYMMDD_FORMAT, YY_DD_FORMAT, YYYYMMDD_FORMAT
)
from glom import glom
from dateutil import parser
from datetime import date

import logging
import json

logger = logging.getLogger(__name__)


class RecordNormalizer:
    """
    Base class for normalizing different record types corresponding classes have been created.
    By default existing record fields will not be changed, unless included in ALLOW_EDIT_FIELDS. If a field
    included in the normalization is not present, it will be added to the record. If under REPEATABLE_FIELDS.
    a new record field is added.

    Args:
        sierra_data: Optionally, can normalize records from SIERRA. Must be in specific format,
        e.g converted with SierraResponseConverter. examples at: tests/sierra/output
        classified_fields: Optionally can include marc fields, will follow the rules of the tag number.
        Useful to send classified data from core.
        entities: List of Full names (str). If included, will use NormLinker to match with normalized records on KATA elastic.
    """

    def __init__(self, linking_results: List[dict] = [], sierra_data: List[dict] = [], classified_fields: List[List[dict]] = [],
                 ALLOW_EDIT_FIELDS: List[str] = ["925"], REPEATABLE_FIELDS: List[str] = ["667"]):

        # Include, if will replace existing field
        self.ALLOW_EDIT_FIELDS = ALLOW_EDIT_FIELDS
        # include, if should be added alongside existing fields
        self.REPEATABLE_FIELDS = REPEATABLE_FIELDS
        # leader applied to new records
        self.DEFAULT_LEADER = "01682nz  a2200349n  4500" # must be 24 digits

    def _setup_records(self, linking_results: List[dict], sierra_data: List[dict], classified_fields: List[List[dict]] = []) -> JSONReader:
        """Setup initial MARC records and data.
        
        If no linked entities or more than one linked entity found, we create a new record.
        If one linked entity found, we create an updated record from the linked entity data.
        """
        linked_records = []
        
        def handle_create_new_record(entity, idx):
            logger.info(f"No linked entities found for {entity}, Creating new record.")
            linked_records.append({
                "leader": self.DEFAULT_LEADER,
                "fields": []
            })
            self.records_extra_data.append({
                "entity": entity,
                "classified_fields": classified_fields[idx] if idx < len(classified_fields) else [],
                "edited": False,
            })
    
        for idx, linked in enumerate(linking_results or []):
            
            if not isinstance(linked, dict):
                logger.debug(f"Skipping invalid linked result: {linked}")
                continue

            entity = linked.get("original_entity")
            linked_info = linked.get("linked_info", [])

            if not isinstance(linked_info, list) or not linked_info:
                # No linked entities found, create new record
                handle_create_new_record(entity, idx)
                continue
            
            elif len(linked_info) > 1:
                # Multiple linked entities found, create new record
                handle_create_new_record(entity, idx)
                continue   

            elif len(linked_info) == 1:
                # one record match found, we update existing record
                
                linked_item = linked_info[0]
                if not isinstance(linked_item, dict):
                    continue
                
                # handle case where we have linked an entity without a record
                if not linked_item.get("json", None):
                    handle_create_new_record(entity, idx)
                    continue
                
                linked_records.append(linked_item.get("json", {}))
                
                self.records_extra_data.append({
                    "entity": entity,
                    "viaf": linked_item.get("viaf", {}),
                    "classified_fields": classified_fields[idx] if idx < len(classified_fields) else [],
                    "edited": True
                })
                continue
        
        self.records_extra_data.extend(
            {
                "sierraID": obj.get("sierraID"),
                "edited": True
            }
            for obj in (sierra_data or [])
            if isinstance(obj, dict)
        )
        
        all_records = linked_records + (sierra_data or [])
        
        return SafeJSONReader(json.dumps(all_records, ensure_ascii=False))

    @staticmethod
    def current_timestamp():
        """6 digit timestamp, format YYMMDD"""
        return datetime.now().strftime(YYMMDD_FORMAT)

    @staticmethod
    def current_yyyy_dd():
        """format of 2025-03"""
        return datetime.now().strftime(YY_DD_FORMAT)

    @staticmethod
    def _is_person_est_nationality(viaf_record: VIAFRecord) -> bool:
        return hasattr(viaf_record, 'nationality') and viaf_record.nationality == "ee"
    
    def get_formatted_dates(self, viaf_record: VIAFRecord) -> str | None:
        """ Get birth and death date in the form 1878-1940. If only birth date is present, return 1878-.
            If no dates, return empty string.
        """
        birth_date = self._extract_year(viaf_record.birth_date)
        death_date = self._extract_year(viaf_record.death_date) if viaf_record.death_date != 0 else ""
        
        if birth_date and death_date:
            return f"{birth_date}-{death_date}"
        elif birth_date:
            return f"{birth_date}-"
        else:
            return None
        
    def _is_nxx(self, field: Field, n: str):
        """ Check if fields tag is in nxx range. """
        return field.tag.startswith(n)

    def get_record_field_or_none(self, record: Record, tag: str) -> Optional[Field]:
        return record.get_fields(tag)[0] if record.get_fields(tag) else None

    def _field_in_record(self, field: Field, record: Record) -> bool:
        """ Check if field exists in record. """
        existing_fields = record.get_fields(field.tag)
        return any(
            field.data == existing_field.data for existing_field in existing_fields)

    def _filter_equivalent_field_not_in_record(self, record: Record, fields: List[Field]) -> bool:
        """ filter out fields, that do not have an equivalent in the record. """
        return filter(lambda field: not self._field_in_record(field, record), fields)
    
    def _extract_year(self, value: str) -> str:
        if value is None:
            return ""

        if isinstance(value, (datetime, date)):
            return str(value.year)

        try:
            dt = parser.parse(str(value), fuzzy=True)
            parsed_year = str(dt.year)
            logger.info(f"Extracted year '{parsed_year}' from value '{value}'")
            return parsed_year
        except Exception as e:
            logger.info(f"Failed to extract year string '{value}': {e}")
            return ""
    
    def _format_date(self, value: str) -> str:
        
        if not value:
            return ""

        if isinstance(value, (datetime, date)):
            return value.strftime(YYYYMMDD_FORMAT)

        val = str(value).strip()

        try:
            dt = parser.parse(val, fuzzy=False, default=datetime(1, 1, 1))
        except Exception:
            return ""

        if len(val) == 4 and val.isdigit():
            return dt.strftime("%Y")  # YYYY
        if len(val) in (6, 7):  # YYYYMM or YYYY-MM
            return dt.strftime("%Y%m")  # YYYYMM
        return dt.strftime(YYYYMMDD_FORMAT)  # YYYYMMDD

    def get_subfield(self, record: Record, tag: str, subfield: str, default: str) -> str:
        """ get record existing subfield value or assign a fallback value. """

        field = self.get_record_field_or_none(record, tag)

        if field is None:
            return default

        subfields = field.get_subfields(subfield)
        return subfields[0] if subfields else default

    def _handle_default_fields(self, record: Record, *fields: List[Field]) -> Record:
        """ Default behavior - add field to record iff not present already """
        record.add_field(
            *filter(lambda field: field.tag not in [
                f.tag for f in record.get_fields()], fields)
        )

    def _handle_editable_fields(self, record: Record, *fields: List[Field]) -> Record:
        """ replace existing field with a new field. """

        editable_fields = filter(
            lambda field: field.tag in self.ALLOW_EDIT_FIELDS, fields)

        tags = [f.tag for f in editable_fields]

        record.remove_fields(
            *tags
        )
        record.add_field(
            *editable_fields
        )

    def _handle_repeatable_fields(self, record: Record, *fields: List[Field]) -> Record:
        """ add field to the record & don't replace existing field."""

        repeatable_fields = [
            field for field in fields if field.tag in self.REPEATABLE_FIELDS]

        record.add_field(
            *repeatable_fields
            # *self._filter_equivalent_field_not_in_record(
            #     record, repeatable_fields)
        )

    def _add_fields_to_record(self, record: Record, fields: List[Field]) -> Record:
        
        cleaned_fields = []
        
        for field in fields:
            
            # Always assume control fields cleaned
            if field.tag < "010" and field.tag.isdigit():
                cleaned_fields.append(field)
                continue
            
            # filter out subfields that are empty or 0 (VIAF returns 0 for unknown dates)
            field.subfields = [
                sub for sub in field.subfields 
                if sub.value and sub.value not in ["0", 0]
            ]
            
            # only keep the field if it still has subfields left
            if field.subfields:
                cleaned_fields.append(field)

        if not cleaned_fields:
            return record
        
        self._handle_repeatable_fields(record, *cleaned_fields)
        self._handle_editable_fields(record, *cleaned_fields)
        self._handle_default_fields(record, *cleaned_fields)
        
        return record

    def _add_author(self, record: Record, viaf_record: VIAFRecord) -> Optional[Field]:
        
        existing_author: Optional[Field] = record.get(
            "100") or record.get("110") or record.get("111")
        if existing_author:
            return record

        type_map = {
            "Personal": "100",
            "Corporate": "110",
            "Collective": "111"
        }

        author_type = viaf_record.name_type
        tag = type_map.get(author_type, "100")

        fields = [
            Field(
                tag=tag,
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("a", viaf_record.name),
                ]
            )
        ]
        
        if viaf_record:
            author_dates = self.get_formatted_dates(viaf_record)
            if author_dates:
                fields[0].add_subfield("d", author_dates)

        self._add_fields_to_record(record, fields)
        
        if viaf_record:
            self._include_name_variations(record, viaf_record)
            
    def _include_name_variations(self, record: Record, viaf_record: VIAFRecord, filter_variations=True) -> None:
        """ Include name variations from VIAF record as 400|t fields """
        
        if not viaf_record or not viaf_record.name_variations:
            return
        
        existing_name_variations = record.get_fields("400")
        existing_variations = [sf.value for field in existing_name_variations for sf in field.get_subfields("a")]
        
        if filter_variations:
            allowed_variations = filter_names(viaf_record.name_variations)
            logger.debug(
                f"filtered out {len(viaf_record.name_variations) - len(allowed_variations)} name variations for '{viaf_record.name}'"
            )
            
        else:
            allowed_variations = viaf_record.name_variations
        
        fields = []
        
        for variation in allowed_variations:
            if variation not in existing_variations:
                fields.append(
                    Field(
                        tag="400",
                        indicators=EMPTY_INDICATORS,
                        subfields=[
                            Subfield("a", variation)
                        ]
                    )
                )
        
        self._add_fields_to_record(record, fields)
        
    def _move680_fields_to_667(self, record: Record) -> None:
        """ Move existing 680 fields to 667, if any. """
        fields_680 = record.get_fields("680")
        if not fields_680:
            return

        fields_667 = [
            Field(
                tag="667",
                indicators=EMPTY_INDICATORS,
                subfields=field.subfields
            ) for field in fields_680
        ]

        record.remove_fields("680")
        self._add_fields_to_record(record, fields_667)
        
    
    def _include_classified_fields(self, record: Record, classified_fields: list[dict]) -> None:
        """Include classified fields from core, if any.
        e.g. classified_fields=[{'670': {'ind1': ' ', 'ind2': '0', 'subfields': [{'a': 'Päikesekiri, 2021'}]}}]
        
        For each record, we need a list of dicts, to handle repeatable fields.
        """
        if not classified_fields:
            return
        
        fields = [
            Field(
                tag=str(tag),
                indicators=v.get("indicators", [v.get("ind1", " "), v.get("ind2", " ")]),
                subfields=[Subfield(code, value) for sub in v.get("subfields", []) for code, value in sub.items()]
            )
            for field_dict in classified_fields
            for tag, v in field_dict.items()
        ]

        self._add_fields_to_record(record, fields)

    def _normalize_common(self, record: Record, is_editing_existing_record: bool, classified_fields: List[dict]) -> None:
        """Common logic for all normalizations. 
        - Includes note about record being created/edited.  
        - include date note with a different subfield, depending on if record is new or edited.
        - move existing 680 fields to 667
        """
        self._include_classified_fields(record, classified_fields)
        
        # before adding new notes
        self._move680_fields_to_667(record)
        
        note = "Muudetud AI poolt" if is_editing_existing_record else "Loodud AI poolt"
        date_note = f"KRATT {self.current_yyyy_dd()}"

        field_667 = Field(tag="667",
                          indicators=EMPTY_INDICATORS,
                          subfields=[Subfield("a", note)])

        fields = [field_667]
        
        if is_editing_existing_record:
            field_925 = Field(tag="925",
                              indicators=EMPTY_INDICATORS,
                              subfields=[
                                  Subfield("p", self.get_subfield(record, "925", "p", date_note))
                              ])
            fields.append(field_925)
            
        else:
            field_925 = Field(tag="925",
                              indicators=EMPTY_INDICATORS,
                              subfields=[
                                  Subfield("t", self.get_subfield(record, "925", "t", date_note))
                              ])
            fields.append(field_925)
        
        self._add_fields_to_record(record, fields)

        return record

    def _get_viaf_search_term(self, record: Record, entity: Optional[str]) -> Optional[str]:
        """ prioritize entity name, if not available, use author name. """
        if entity:
            return entity

        author_field = record.get("100") or record.get("110") or record.get("111")
        if author_field:
            return author_field.get_subfields("a")[0] if author_field.get_subfields("a") else None

        logger.warning(
            "No entity or author name found for VIAF search. Skipping VIAF enrichment.")

    def _get_viaf_record(self, record: Record, viaf_id: Optional[int] = None,
            entity: Optional[str] = None, viaf_field: str = DEFAULT_VIAF_FIELD,
            threshold: float = VIAF_SIMILARITY_THRESHOLD, verify: bool = VERIFY_VIAF_RECORD,
            max_records: int = MAX_VIAF_RECORDS_TO_VERIFY
    ) -> Optional[VIAFRecord]:
        viaf_record = None

        try:
            viaf_client = VIAFClient()

            if viaf_id:
                viaf_records = viaf_client.get_normalized_data_by_ids([viaf_id])
                if viaf_records:
                    viaf_record = viaf_records[0]
            else:
                search_term = self._get_viaf_search_term(record, entity)

                if search_term:
                    logger.info(
                        f"Searching for VIAF record with search term: {search_term}")

                    if not verify:
                        logger.warning(
                            f"Record verification is turned off. If multiple records are " \
                            f"detected for search term '{search_term}', the first " \
                            f"result is automatically returned. This might lead to " \
                            f"some inaccuracies!"
                        )

                    viaf_record = viaf_client.get_normalized_data_by_search_term(
                        search_term=search_term,
                        field=viaf_field,
                        max_records=max_records,
                        verify=verify,
                        threshold=threshold
                    )
                    if viaf_record:
                        logger.debug(f"VIAF {search_term}, linked to ID: {viaf_record.viaf_id}")

        except Exception as e:
            logger.error(
                f"Error fetching VIAF record with ID={viaf_id} / entity='{entity}': {e}"
            )
        return viaf_record

    def _normalize_record(self, record: Record, sierraID: str,
                          viaf_record: VIAFRecord, is_editing_existing_record: bool, 
                          original_entity: str) -> Record:
        return record
    
    def get_record(self, index: int) -> Record:
        """Get normalized record by index."""
        for idx, record in enumerate(self):
            if idx == index:
                return record
        raise IndexError("Record index out of range.")

    @property
    def data(self) -> List[dict]:
        """Shorthand to get all normalized records as dict, skipping failures."""
        result = []
        for record in self:
            try:
                result.append(record.as_dict())
            except Exception as e:
                logger.error(f"Failed to normalize record: {e}")
                continue
        return result
    
    @property
    def first(self) -> Record:
        return next(iter(self))
      
    def __iter__(self) -> Iterator:
        # viaf_id_path = "viaf.original.queryResult.viafID"
        viaf_id_path = "viaf.parsed.viaf_id"
        
        sierra_id_path = "sierraID"
        
        for record, extra_data in zip(self.records, self.records_extra_data):
            
            sierra_id = glom(extra_data, sierra_id_path, default="")
            viaf_id = glom(extra_data, viaf_id_path, default=None)
            
            classified_fields = extra_data.get("classified_fields", [])
            entity = extra_data.get("entity")
            is_editing_existing_record = extra_data.get("edited") == True

            viaf_record = self._get_viaf_record(record, viaf_id, entity)
            if viaf_record:
                logger.debug(
                    f"linked VIAF record with ID {viaf_record.viaf_id} for entity '{entity}'"
                )
            
            record = self._normalize_common(record, is_editing_existing_record, classified_fields)

            normalized_record = self._normalize_record(
                record, sierra_id, viaf_record, is_editing_existing_record, original_entity=entity)

            normalized_record.fields.sort(key=lambda field: field.tag)

            yield normalized_record



from rara_tools.normalizers.viaf import VIAFRecord, VIAFClient


def test_fetch_clusters_by_id_list():
    viaf_ids = ["7432247", "456"]
    client = VIAFClient()

    results = client.fetch_viaf_clusters(viaf_ids)
    assert len(results) == 2
    assert results["456"] == {}
    assert len(results["7432247"]) > 0


def test_fetch_viaf_results_for_normalizer():
    viaf_ids = ["7432247", "456"]
    client = VIAFClient()

    results = client.get_normalized_data_by_ids(viaf_ids)
    assert len(results) == 2

def test_get_normalized_data_by_search_term():
    client = VIAFClient()
    entity = "Kaja Kallas"

    # Test without verification
    record = client.get_normalized_data_by_search_term(
        search_term=entity,
        verify=False
    )
    assert record.name == "Kallas, Siim"

    record = client.get_normalized_data_by_search_term(
        search_term=entity,
        verify=True
    )
    assert record.name == "Kallas, Kaja"

    record = client.get_normalized_data_by_search_term(
        search_term = "ahsjhajhsjh",
        verify=True,
        max_records=5
    )
    assert record == None

def test_viaf_search_term_query_working():
    client = VIAFClient()
    assert client.check_search_term_query()

def test_viaf_id_query_working():
    client = VIAFClient()
    assert client.check_id_query()


def test_subfield_based_main_field_extraction():
    """ VIAFRecod.name should be retrieved from
    subfield (a) for persons, corporations etc, but
    from field (t) for titles.
    """
    client = VIAFClient()
    record = client.get_normalized_data_by_search_term(
        search_term="Kevade",
        field="local.uniformTitleWorks",
        verify=True
    )
    assert record.name == "Kevade"

    record = client.get_normalized_data_by_search_term(
        search_term="Oskar Luts",
        verify=True
    )
    assert record.name == "Luts, Oskar"

def test_changing_allowed_sources():
    client = VIAFClient(allowed_viaf_sources=["PLWABN"])
    record = client.get_normalized_data_by_search_term(
        search_term="Anora",
        field="local.uniformTitleWorks",
        verify=False
    )
    assert record.name == "Anora (film)"

    client = VIAFClient(allowed_viaf_sources=["LC"])
    record = client.get_normalized_data_by_search_term(
        search_term="Anora",
        field="local.uniformTitleWorks",
        verify=False
    )
    assert record.name == "Anora (Motion picture)"

def test_int_entities_pass():
    client = VIAFClient()
    record = client.get_normalized_data_by_search_term(
        search_term="Prosper Merimee",
        field="local.uniformTitleWorks",
        verify=True
    )
    # Make sure that no error is thrown, while accessing the record
    assert record == None

def test_accent_removal():
    pairs = [
        ("Prosper Merimee", "Prosper Mérimée"),
        ("Ernst Öpik", "Ernst Opik"),
        ("Fållan", "Fallan"),
        ("Õhtuõpik", "Ohtuopik"),
        ("Ööäär", "Ooaar"),
        ("Brontë", "Bronte"),
        ("Iñes", "Ines"),
        ("Håkon Søren", "Hakon Soren"),
        ("Łukasz", "Lukasz"),
        ("Côte d’Azur", "Cote d’Azur"),
        ("České Budějovice", "Ceske Budejovice")
    ]

    for p1, p2 in pairs:
        s1 = VIAFClient.normalize_latin(p1)
        s2 = VIAFClient.normalize_latin(p2)
        assert s1 == s2



import pytest
import os
from pprint import pprint
from rara_tools.core_formatters.core_formatter import format_keywords, format_meta, get_udk072, get_udk080
from rara_tools.constants.meta_extractor import PUBLISHER_KEY
from tests.test_utils import read_json_file

ROOT_DIR = os.path.join("tests", "test_data", "formatter")
INPUT_KEYWORDS_FILE_PATHS = [
    os.path.join(ROOT_DIR, "keywords_1.json"),
    os.path.join(ROOT_DIR, "keywords_2.json"),
    os.path.join(ROOT_DIR, "keywords_3.json")
]
INPUT_META_FILE_PATHS = [
    os.path.join(ROOT_DIR, "epub_meta.json"),
    os.path.join(ROOT_DIR, "mets_alto_meta.json"),
    os.path.join(ROOT_DIR, "pdf_meta_2.json"),
    os.path.join(ROOT_DIR, "pdf_meta.json")
]

INPUT_KEYWORDS = [
    read_json_file(keyword_file_path)
    for keyword_file_path in INPUT_KEYWORDS_FILE_PATHS
]

INPUT_META_DICTS = [
    read_json_file(meta_file_path)
    for meta_file_path in INPUT_META_FILE_PATHS
]

def test_formatting_keywords_for_core():
    for keyword_dict_list in INPUT_KEYWORDS:
        formatted_keywords = format_keywords(keyword_dict_list)
        #pprint(formatted_keywords)
        assert formatted_keywords
        assert isinstance(formatted_keywords, list)


def test_formatting_meta_for_core():
    for meta_dict in INPUT_META_DICTS:
        formatted_meta = format_meta(meta_dict)
        #pprint(formatted_meta)
        assert formatted_meta
        assert isinstance(formatted_meta, dict)


def test_validating_keywords():
    keyword_dict_list = INPUT_KEYWORDS[-1]
    assert len(keyword_dict_list) == 6
    formatted_keywords = format_keywords(keyword_dict_list)
    assert len(formatted_keywords) == 2

def test_removing_publisher_from_authors():
    meta_dict = INPUT_META_DICTS[-1]
    formatted_meta = format_meta(meta_dict)
    assert formatted_meta["meta"]["publisher"] == "s.n"
    for author in formatted_meta["meta"]["authors"]:
        assert author.get("author_role") != PUBLISHER_KEY
    assert isinstance(formatted_meta, dict)

def test_title_key_without_authors():
    meta_dict = INPUT_META_DICTS[2]
    formatted_meta = format_meta(meta_dict)
    for title in formatted_meta["meta"]["titles"]:
        assert title.get("title_type_int") == 130
    assert isinstance(formatted_meta, dict)

def test_all_authors_have_types():
    for meta_dict in INPUT_META_DICTS:
        formatted_meta = format_meta(meta_dict)
        authors = formatted_meta.get("authors", [])
        for author in authors:
            assert author.get("primary_author_type")

def test_getting_udk072():
    keyword_dict_list = INPUT_KEYWORDS[-1]
    assert len(keyword_dict_list) == 6
    udk072_list = get_udk072(keyword_dict_list)
    assert len(udk072_list) == 1

def test_getting_udk080():
    keyword_dict_list = INPUT_KEYWORDS[-1]
    assert len(keyword_dict_list) == 6
    udk080_list = get_udk080(keyword_dict_list)
    assert len(udk080_list) == 1



from rara_tools.constants.linker import (
    LOGGER, URLSource, KeywordSource, EntityType, KeywordType,
    KEYWORD_MARC_MAP,  KEYWORD_TYPES_TO_IGNORE, KEYWORD_TYPE_MAP,
    EMS_ENTITY_TYPES, SIERRA_ENTITY_TYPES, UNLINKED_KEYWORD_MARC_FIELD,
    URL_SOURCE_MAP
)

from rara_tools.parsers.marc_records.person_record import PersonRecord
from rara_tools.parsers.marc_records.organization_record import OrganizationRecord
from rara_tools.parsers.marc_records.title_record import TitleRecord
from rara_tools.utils import format_date
from typing import List, Dict, NoReturn, Tuple, Any


class FormattedObject:
    def __init__(self, object_dict: dict, linked_doc: Any, original_entity_key: str):
        self.object_dict: dict = object_dict
        self.linked_doc: Any = linked_doc
        self.viaf_info: dict = self.linked_doc.viaf.get("parsed", {}) if self.linked_doc else {}
        self.original_entity: str = self.object_dict.get(original_entity_key)
        self.is_linked: bool = True if self.linked_doc else False

        self.__original_record: PersonRecord | OrganizationRecord | TitleRecord | None = None
        self.__persons_title: str | None = None
        self.__dates: str | None = None
        self.__numeration: str | None = None
        self.__location: str | None = None
        self.__organization_sub_unit: str | None = None
        self.__entity: str | None = None
        self.__titles: str | None = None
        self.__identifier: str | None = ""


    @property
    def original_record(self) -> PersonRecord | OrganizationRecord | None:
        if not self.__original_record and self.linked_doc and self.linked_doc.json:
            try:
                if self.entity_type == EntityType.PER:
                    original_record = PersonRecord(self.linked_doc.json)
                elif self.entity_type == EntityType.ORG:
                    original_record = OrganizationRecord(self.linked_doc.json)
                elif self.entity_type == EntityType.TITLE:
                    original_record = TitleRecord(self.linked_doc.json)
                else:
                    original_record = None
            except Exception as e:
                LOGGER.exception(
                    f"Could not retrieve JSON from LinkedDoc instance. Exception: '{e}'."
                )
                original_record = None
            self.__original_record = original_record
        return self.__original_record


    @property
    def entity(self) -> str:
        if self.__entity == None:
            if self.linked_doc != None:
                if self.entity_type == EntityType.ORG and self.original_record:
                    self.__entity = self.original_record.original_name.get("a", "")
                else:
                    self.__entity = self.linked_doc.linked_entity
                if not self.__entity and self.viaf_info:
                    self.__entity = self.viaf_info.get("name", self.original_entity)
            else:
                self.__entity = self.original_entity
        return self.__entity



    @property
    def dates(self) -> str:
        if self.__dates == None:
            self.__dates = ""
            if self.viaf_info:
                birth_date = format_date(self.viaf_info.get("birth_date", ""))
                death_date = format_date(self.viaf_info.get("death_date", ""))
                if not death_date:
                    death_date = ""

                if birth_date:
                    self.__dates = f"{birth_date}-{death_date}"

            if self.original_record and not self.__dates:
                if self.entity_type == EntityType.PER:
                    self.__dates = self.original_record.life_years
                elif self.entity_type == EntityType.ORG:
                    self.__dates = self.original_record.dates
                elif self.entity_type == EntityType.TITLE:
                    self.__dates = self.original_record.author_life_years

        return self.__dates


    @property
    def numeration(self) -> str:
        if self.__numeration == None:
            self.__numeration = ""
            if self.original_record:
                if self.entity_type == EntityType.PER:
                    self.__numeration = self.original_record.original_name.get("b", "")
                elif self.entity_type == EntityType.ORG:
                    self.__numeration = self.original_record.numeration
        return self.__numeration

    @property
    def location(self) -> str:
        if self.__location == None:
            self.__location = ""
            if self.entity_type == EntityType.ORG and self.original_record:
                self.__location = self.original_record.location
        return self.__location

    @property
    def organisation_sub_unit(self) -> str:
        if self.__organization_sub_unit == None:
            self.__organization_sub_unit = ""
            if self.entity_type == EntityType.ORG and self.original_record:
                self.__organization_sub_unit = self.original_record.original_name.get("b", "")
        return self.__organization_sub_unit

    @property
    def titles(self) -> str:
        if self.__titles == None:
            if self.entity_type == EntityType.PER and self.original_record:
                self.__titles = self.original_record.name_specification
            else:
                self.__titles = ""
        return self.__titles

    @property
    def identifier(self) -> str:
        if self.__identifier == None:
            self.__identifier = ""
            if self.original_record:
                self.__identifier = self.original_record.identifier
        return self.__identifier



from rara_tools.constants import EMPTY_INDICATORS
from rara_tools.normalizers.viaf import VIAFRecord

from rara_tools.normalizers import RecordNormalizer

from pymarc import Field, Subfield, Record
from typing import List


class AuthoritiesRecordNormalizer(RecordNormalizer):
    """ Normalize authorities records """

    def __init__(self, linking_results: List[dict] = [], sierra_data: List[dict] = [],
                 classified_fields: List[List[dict]] = [],
                 ALLOW_EDIT_FIELDS: List[str] = ["008", "925"],
                 REPEATABLE_FIELDS: List[str] = ["024", "035", "400", "667"]):

        super().__init__(linking_results, sierra_data, classified_fields)
        self.ALLOW_EDIT_FIELDS = ALLOW_EDIT_FIELDS
        self.REPEATABLE_FIELDS = REPEATABLE_FIELDS
        self.records_extra_data = []
        self.sierra_data = sierra_data
        self.records = self._setup_records(linking_results, sierra_data, classified_fields)

    def _normalize_sierra(self, record: Record, sierraID: str, is_editing_existing_record: bool) -> Record:
        """008 updated only for new records, unless editing where prefix is preserved."""

        suffix_008 = "|n|adnnnaabn          || |a|      "

        if is_editing_existing_record:
            # Try to reuse prefix from existing 008 field if present
            existing_008 = next((f for f in record.fields if f.tag == "008" and hasattr(f, "data")), None)
            if existing_008 and len(existing_008.data) >= 6:
                prefix = existing_008.data[:6]
            else:
                prefix = self.current_timestamp()  # fallback if missing
        else:
            prefix = self.current_timestamp()

        fields = [
            Field(
                tag="008",
                data=f"{prefix}{suffix_008}"
            )
        ]

        field_040 = Field(
            tag="040",
            indicators=EMPTY_INDICATORS,
            subfields=[
                Subfield("a", self.get_subfield(record, "040", "a", "ErESTER")),
                Subfield("b", self.get_subfield(record, "040", "b", "est")),
                Subfield("c", self.get_subfield(record, "040", "c", "ErEster")),
            ]
        )
        fields.append(field_040)

        self._add_fields_to_record(record, fields)
        
        return record
    
    def _add_birth_and_death_dates(self, record: Record, viaf_record: VIAFRecord) -> None:
        
        formatted_birth_date = self._format_date(viaf_record.birth_date)
        formatted_death_date = self._format_date(viaf_record.death_date) if viaf_record.death_date != 0 else ""

        birth_date = self.get_subfield(
            record, "046", "f", formatted_birth_date)
        death_date = self.get_subfield(
            record, "046", "g", formatted_death_date)
        
        if not birth_date and not death_date:
            return

        subfields_046 = [
            Subfield("f", birth_date),
            Subfield("g", death_date),
        ]

        self._add_fields_to_record(
            record, [Field(tag="046", indicators=EMPTY_INDICATORS, subfields=subfields_046)])

    def _add_viaf_url_or_isni(self, record: Record, viaf_record: VIAFRecord) -> None:
        viaf_url = viaf_record.viaf_url

        subfields = [Subfield("0", self.get_subfield(
            record, "024", "0", viaf_url))]

        if viaf_record.has_isni:
            subfields.append(Subfield("2", "isni"))

        field = Field(tag="024", indicators=EMPTY_INDICATORS,
                      subfields=subfields)

        self._add_fields_to_record(record, [field])

    def _add_nationality(self, record: Record, viaf_record: VIAFRecord) -> None:
        """ Non-repeatable field 043 - adds ee only if is estonian nationality and 
        the records does not have the field already."""
        
        is_person_est = self._is_person_est_nationality(viaf_record)
        
        if is_person_est:
            fields = [
                Field(
                    tag="043",
                    indicators=EMPTY_INDICATORS,
                    subfields=[Subfield("c", "ee")])
                ]

            self._add_fields_to_record(record, fields)

    def _normalize_viaf(self, record: Record, viaf_record: VIAFRecord) -> None:
        """
        Attempts to enrich the record with VIAF data.

        024 - repeatable field, add VIAF URL to subfield 0. If ISNI found, add to subfield 2
        043 - repeatable field. Add "ee" if found to be estonian nationality
        046 - non-repeatable field, add birth and death dates
        100, 110, 111 - non-repeatable field, attempts to add author type, if missing.

        """
        if not viaf_record:
            return

        self._add_nationality(record, viaf_record)
        self._add_viaf_url_or_isni(record, viaf_record)
        self._add_birth_and_death_dates(record, viaf_record)
        self._add_author(record, viaf_record)

    def _normalize_record(self, record: Record, sierraID: str,
                          viaf_record: VIAFRecord, 
                          is_editing_existing_record: bool,
                          original_entity: str) -> Record:
        
        self._normalize_sierra(record, sierraID, is_editing_existing_record)
        self._normalize_viaf(record, viaf_record)

        return record



import logging

from rara_tools.constants.normalizers import EntityType

COMPONENT_KEY = "linker"


class Tasks:
    BASE = "base_linker_task"
    VECTORIZE = "vectorize_text"
    VECTORIZE_WITH_CORE = "vectorize_text_with_core_logic"
    PIPELINE = "link_keywords_with_core_logic"

    LINK_AND_NORMALIZE = "core_linker_with_normalization"
    VECTORIZE_AND_INDEX = "core_vectorize_and_index"
    RECEIVE_LINK_AND_NORMALIZE = "receive_link_and_normalize"


class Queue:
    LINKER = "linker"
    VECTORIZER = "vectorizer"


class StatusKeys:
    VECTORIZE_CONTEXT = "vectorize_context"
    LINK_KEYWORDS = "link_keywords"


class URLSource:
    VIAF = "VIAF"
    SIERRA = "Sierra"
    EMS = "EMS"


class KeywordType:
    LOC = "Kohamärksõnad"
    TIME = "Ajamärksõnad"
    TOPIC = "Teemamärksõnad"
    GENRE = "Vormimärksõnad"
    TITLE = "Teose pealkiri"
    PER = "Isikunimi"
    ORG = "Kollektiivi nimi"
    EVENT = "Ajutine kollektiiv või sündmus"
    CATEGORY = "Valdkonnamärksõnad"
    UDC = "UDC Summary"
    UDK = "UDK Rahvusbibliograafia"


class KeywordMARC:
    PER = 600
    ORG = 610
    TOPIC = 650
    GENRE = 655
    TIME = 648
    LOC = 651
    EVENT = 611
    TITLE = 630
    TITLE_LINKED = 600


class KeywordSource:
    EMS = "EMS"
    SIERRA = "SIERRA"
    VIAF = "VIAF"
    AI = "AI"


class Filters:
    AUTHOR = "author"
    YEAR = "year"


class Error:
    VECTORIZATION = "Failed to vectorize text!"
    LINKING_KEYWORDS = "Failed to link keywords!"
    LINKING_META = "Failed to link meta!"

UNLINKED_KEYWORD_MARC_FIELD = 693

ALLOWED_FILTERS_MAP = {
    EntityType.PER: [Filters.YEAR],
    EntityType.ORG: [Filters.YEAR],
    EntityType.TITLE: [Filters.YEAR, Filters.AUTHOR],
    EntityType.KEYWORD: [],
    EntityType.LOC: []
}
KEYWORD_MARC_MAP = {
    KeywordType.LOC: KeywordMARC.LOC,
    KeywordType.TIME: KeywordMARC.TIME,
    KeywordType.TOPIC: KeywordMARC.TOPIC,
    KeywordType.GENRE: KeywordMARC.GENRE,
    KeywordType.TITLE: KeywordMARC.TITLE,
    KeywordType.ORG: KeywordMARC.ORG,
    KeywordType.PER: KeywordMARC.PER,
    KeywordType.EVENT: KeywordMARC.EVENT
}

URL_SOURCE_MAP = {
    EntityType.PER: URLSource.VIAF,
    EntityType.ORG: URLSource.VIAF,
    EntityType.TITLE: URLSource.VIAF,
    EntityType.KEYWORD: URLSource.EMS,
    EntityType.LOC: URLSource.EMS
}

# Ignore those "keyword types" while linking the
# rara-subject-indexer results
KEYWORD_TYPES_TO_IGNORE = [
    KeywordType.CATEGORY,
    KeywordType.UDC,
    KeywordType.UDK
]

ALLOWED_ENTITY_TYPES = [
    EntityType.PER,
    EntityType.ORG,
    EntityType.KEYWORD,
    EntityType.LOC,
    EntityType.TITLE,
    EntityType.UNK,
]

KEYWORD_TYPE_MAP = {
    KeywordType.TIME: EntityType.KEYWORD,
    KeywordType.GENRE: EntityType.KEYWORD,
    KeywordType.LOC: EntityType.LOC,
    KeywordType.PER: EntityType.PER,
    KeywordType.ORG: EntityType.ORG,
    KeywordType.TOPIC: EntityType.KEYWORD,
    KeywordType.TITLE: EntityType.TITLE,
    KeywordType.EVENT: EntityType.ORG
}

EMS_ENTITY_TYPES = [EntityType.KEYWORD, EntityType.LOC]
SIERRA_ENTITY_TYPES = [EntityType.PER, EntityType.ORG, EntityType.TITLE]
VIAF_ENTITY_TYPES = [EntityType.PER, EntityType.ORG, EntityType.TITLE]

# Params for filters
MIN_AUTHOR_SIMILARITY = 0.95
YEAR_EXCEPTION_VALUE = True

LOGGER_NAME = "rara-tools-norm-linker"
LOGGER = logging.getLogger(LOGGER_NAME)

MAIN_TAXONOMY_LANG = "et"



import json
import pytest
import os
import sys

from rara_tools.digar_schema_converter import DIGARSchemaConverter

def load_json(file_path: str):
    with open(file_path, "r") as f:
        data = json.load(f)
    return data


TEST_DIGITIZER_OUTPUT_FILE = os.path.join(".", "tests", "test_data", "b1267058_test_digitizer_output.json")
TEST_DIGITIZER_OUTPUT_FILE_2 = os.path.join(".", "tests", "test_data", "b5493797_test_digitizer_output_shortened_empty_image_pages.json")

TEST_DIGITIZER_OUTPUT = load_json(TEST_DIGITIZER_OUTPUT_FILE)
TEST_DIGITIZER_OUTPUT_2 = load_json(TEST_DIGITIZER_OUTPUT_FILE_2)

TEST_SIERRA_ID = "b1267058"
TEST_GENERATED_ID = "hsasaHSAHHGDhb"
TEST_PERMALINK = "https://www.digar.ee/b1267058"

TEST_SIERRA_ID_2 = "b5493797"
TEST_GENERATED_ID_2 = "mzmudhju38hd3ndlk"
TEST_PERMALINK_2 = "https://www.digar.ee/b5493797"

def test_digar_schema_converstion_default():
    converter = DIGARSchemaConverter(
        digitizer_output=TEST_DIGITIZER_OUTPUT,
        sierra_id=TEST_SIERRA_ID,
        generated_id=TEST_GENERATED_ID
    )
    digar_schema = converter.digar_schema

    # check that all neseccary fields are present
    assert "dc:language" in digar_schema
    assert "dcterms:provenance" in digar_schema
    assert "dc:identifier" in digar_schema
    assert "dcterms:hasPart" in digar_schema
    assert "dcterms:conformsTo" in digar_schema

    languages = [lang.get("value") for lang in digar_schema.get("dc:language")]
    # check that languages are converted into ISO-693-2
    for lang in languages:
        assert len(lang) == 3


    # check that ratio is converted into percentage
    text_quality = digar_schema.get("dcterms:conformsTo")[0].get("value")
    assert isinstance(text_quality, str)


def test_digar_schema_id_generation():
    """ Tests ID generation logic.
    """
    converter = DIGARSchemaConverter(
        digitizer_output=TEST_DIGITIZER_OUTPUT,
        sierra_id=TEST_SIERRA_ID,
        generated_id=TEST_GENERATED_ID,
        permalink=TEST_PERMALINK

    )

    #If permalink is given, this should be used as base ID
    digar_schema = converter.digar_schema
    first_segment_id = digar_schema["dcterms:hasPart"][0]["dcterms:hasPart"][0]["@id"]

    assert first_segment_id.startswith(TEST_PERMALINK)

    converter = DIGARSchemaConverter(
        digitizer_output=TEST_DIGITIZER_OUTPUT,
        sierra_id=TEST_SIERRA_ID,
        generated_id=TEST_GENERATED_ID
    )

    #If permalink is NOT given, Sierra ID should be used as base ID
    digar_schema = converter.digar_schema
    first_segment_id = digar_schema["dcterms:hasPart"][0]["dcterms:hasPart"][0]["@id"]
    assert first_segment_id.startswith(TEST_SIERRA_ID)


    converter = DIGARSchemaConverter(
        digitizer_output=TEST_DIGITIZER_OUTPUT,
        generated_id=TEST_GENERATED_ID
    )

    #If neiter permalink nor Sierra ID is given, generated ID should be used as base ID
    digar_schema = converter.digar_schema
    first_segment_id = digar_schema["dcterms:hasPart"][0]["dcterms:hasPart"][0]["@id"]
    assert first_segment_id.startswith(TEST_GENERATED_ID)


def test_restricting_languages_with_ratio():
    """ Checks that param `min_language_ratio` influences
    the number of output languages.
    """
    converter = DIGARSchemaConverter(
        digitizer_output=TEST_DIGITIZER_OUTPUT,
        sierra_id=TEST_SIERRA_ID,
        generated_id=TEST_GENERATED_ID,
        permalink=TEST_PERMALINK,
        min_language_ratio=0

    )

    #If permalink is given, this should be used as base ID
    digar_schema = converter.digar_schema
    languages = [lang.get("value") for lang in digar_schema.get("dc:language")]
    assert len(languages) == 7

    converter = DIGARSchemaConverter(
        digitizer_output=TEST_DIGITIZER_OUTPUT,
        sierra_id=TEST_SIERRA_ID,
        generated_id=TEST_GENERATED_ID,
        permalink=TEST_PERMALINK,
        min_language_ratio=0.02

    )

    #If permalink is given, this should be used as base ID
    digar_schema = converter.digar_schema
    languages = [lang.get("value") for lang in digar_schema.get("dc:language")]
    assert len(languages) == 2

    converter = DIGARSchemaConverter(
        digitizer_output=TEST_DIGITIZER_OUTPUT,
        sierra_id=TEST_SIERRA_ID,
        generated_id=TEST_GENERATED_ID,
        permalink=TEST_PERMALINK,
        min_language_ratio=0.5

    )

    #If permalink is given, this should be used as base ID
    digar_schema = converter.digar_schema
    languages = [lang.get("value") for lang in digar_schema.get("dc:language")]
    assert len(languages) == 1


def test_digar_schema_converstion_with_missing_image_pages():
    converter = DIGARSchemaConverter(
        digitizer_output=TEST_DIGITIZER_OUTPUT_2,
        sierra_id=TEST_SIERRA_ID_2,
        generated_id=TEST_GENERATED_ID_2
    )
    digar_schema = converter.digar_schema

    # check that all neseccary fields are present
    assert "dc:language" in digar_schema
    assert "dcterms:provenance" in digar_schema
    assert "dc:identifier" in digar_schema
    assert "dcterms:hasPart" in digar_schema
    assert "dcterms:conformsTo" in digar_schema

    languages = [lang.get("value") for lang in digar_schema.get("dc:language")]
    # check that languages are converted into ISO-693-2
    for lang in languages:
        assert len(lang) == 3


    # check that ratio is converted into percentage
    text_quality = digar_schema.get("dcterms:conformsTo")[0].get("value")
    assert isinstance(text_quality, str)



from rara_tools.parsers.tools.validators import filter_names 
import pytest

are_equal = lambda x, y: not bool(set(x).difference(set(y)))

names_to_validate = [
    "ליסט, פראנץ",
    "Liszt, Franz",
    "Lißt, Franz",
    "ליסט, פרנץ",
    "Liszt, Ferencz",
    "Лист, Франц",
    "Listz",
    "Lißzt, Franz",
    "Lists, Francis",
    "List, Ferenc",
    "List, Frants리스",
    "List, Ferents",
    "李斯特，弗朗西斯庫斯",
    "ᓕᔅᑦ, ᕗᕌᓐᓯᔅᑲᔅ",
    "리스트, 프란치스코"
]

valid_names_1 = [
    "Liszt, Franz",
    "Lißt, Franz",
    "Liszt, Ferencz",
    "Лист, Франц",
    "Listz",
    "Lißzt, Franz",
    "Lists, Francis",
    "List, Ferenc",
    "List, Frants리스",
    "List, Ferents"
]

valid_names_2 = [
    "Liszt, Franz",
    "Lißt, Franz",
    "Liszt, Ferencz",
    "Listz",
    "Lißzt, Franz",
    "Lists, Francis",
    "List, Ferenc",
    "List, Frants리스",
    "List, Ferents"
]

def test_filtering_latin_cyrillic():
    filtered_names = filter_names(names_to_validate, allow_cyrillic=True)
    assert are_equal(filtered_names, valid_names_1)

def test_filtering_latin():
    filtered_names = filter_names(names_to_validate, allow_cyrillic=False)
    assert are_equal(filtered_names, valid_names_2)
